--[[
开放式自定义 OpenAPI 接口

兼容 XXTouchNG 接口实现

本文件仅作为参考，请不要修改本文件
本文件会在重装、更新时被原版覆盖
如果要新增 OpenAPI 接口可参考本文件创建名为 custom-http-api.lua 的文件来实现
例如可将自定义接口部分写在下面文件中
/var/mobile/Media/1ferver/bin/custom-http-api.lua
--]]

local json = require('cjson.safe')
local lfs = require('lfs')
local noexecute = require('no_os_execute')
local archive = require('archive')
local posix = require('posix')
local path_manager = require('path')

function tar_extract(tar_path, to_path, uid, gid, mode)
	local arfh, err = io.open(tar_path, 'r')
	if not arfh then
		return false, err
	end
	if not to_path then
		to_path = path_manager.ensure_dir_end(path_manager.splitext(tar_path))
		sys.mkdir_p(to_path)
	end
    uid = uid or 501
    gid = gid or 501
    mode = mode or 755
	local ar = archive.read{
		format = 'tar',
		reader = function(reader)
			return arfh:read(1024 * 1024)
		end,
	}
	for h in ar.next_header, ar do
		local fname = path_manager.basename(h:pathname())
		if not (fname == '.DS_Store' or fname:sub(1, 2) == '._') then
			local filepath = path_manager.join(to_path, h:pathname())
			if posix.S_ISDIR(h:mode()) == 1 then
				sys.mkdir_p(filepath)
			else
				local dir = path_manager.dirname(filepath)
				if not path_manager.isdir(dir) then
					sys.mkdir_p(dir)
				end
				local f, err = io.open(filepath, 'w')
				if not f then
					return false, err
				end
				for s in ar.data, ar do
					f:write(s)
				end
				f:close()
			end
		end
	end
	noexecute.lchownmod_r(to_path, uid, gid, mode)
	return true, to_path
end

function file_lastline(filename, last)
    last = tonumber(last) or 1
    last = math.floor(last)
    last = (last > 0 and last) or 1
    if last == 1 then
        local f = io.open(filename, 'r')
        if f then
            local s = f:read('*a')
            f:close()
            local t = s:split('\n')
            if t[#t] == '' or t[#t] == '\r' then
                table.remove(t, #t)
            end
            return t[#t]
        end
    else
        local f = io.open(filename, 'r')
        if f then
            local s = f:read('*a')
            f:close()
            local t = s:split('\n')
            if t[#t] == '' or t[#t] == '\r' then
                table.remove(t, #t)
            end
            if #t <= last then
                return table.concat(t, '\r\n')
            else
                return table.concat(t, '\r\n', #t - last, #t)
            end
        end
    end
    return ''
end

if not lua_is_running then
    function lua_is_running()
        local c, h, r = xxtouch.post('/is_running')
        if c == 200 then
            return (json.decode(r) or {code = 0}).code ~= 0
        else
            return false
        end
    end
end

local function _internal_error_message(ctx, code, msg)
    ctx.status = code
    ctx:header('Content-type', 'application/json; charset=utf-8')
    -- sys.log(string.format([[{"code": %d, "message": %q}]], code, tostring(msg)).." "..json.encode({uri = ctx.uri, method = ctx.request_method, headers = ctx.http_headers, query = ctx.query}))
    return string.format([[{"code": %d, "message": %q}]], code, tostring(msg))
end

local function _request_error_message(ctx, code, msg)
    ctx.status = code
    ctx:header('Content-type', 'application/json; charset=utf-8')
    -- sys.log(string.format([[{"code": %d, "message": %q}]], code, tostring(msg)).." "..json.encode({uri = ctx.uri, method = ctx.request_method, headers = ctx.http_headers, query = ctx.query}))
    return string.format([[{"code": %d, "message": %q}]], code, tostring(msg))
end

local function _sh_escape(path)
    if jbroot('/rootfs') == '/' then
        path = rootfs(path)
    end
	path = string.gsub(path, "([ \\()<>'\"`#&*;?~$|])", "\\%1")
	return path
end

local function _read_conf()
	local conf
	local f = io.open(XXT_CONF_FILE_NAME, 'r')
	if f then
		conf = json.decode(f:read('*a')) or {}
		f:close()
	end
	conf = (type(conf) == 'table') and conf or {}
	return conf
end

local function _write_conf(conf)
	local f = io.open(XXT_CONF_FILE_NAME, 'w')
	if f then
        local conf_str = json.encode(conf)
        if type(conf_str) == 'string' then
            f:write(conf_str)
            f:close()
            return true
        else
            f:close()
            return false
        end
    else
        return false
	end
end

local function _number_range(num, min, max, default)
    default = tonumber(default) or 0
    num = tonumber(num) or default
    num = (num <= max) and num or max
    num = (num >= min) and num or min
    return num
end

local nLog_imp_str = [[
do
    local select = select
    local table = {deep_dump = table.deep_dump, concat = table.concat}
    local sys = {log = sys.log}
    local tostring = tostring
    local print = print
    local type = type
    function nLog(...)
    	local outt = {}
        local argc = select("#", ...)
        if argc > 0 then
            for i = 1, argc do
                local arg = select(i, ...)
                if type(arg) == "table" then
                    outt[#outt + 1] = table.deep_dump(arg)
                else
                    outt[#outt + 1] = tostring(arg)
                end
            end
            sys.log(table.concat(outt, '\t'))
        end
    end
end
]]

nLog_imp_str = nLog_imp_str:gsub("\r", ""):gsub("\n", ";")

local proc_operations = dofile(XXT_BIN_PATH..'/module-proc-operations.lua')
local proc_value_operations = proc_operations.proc_value_operations
local proc_queue_operations = proc_operations.proc_queue_operations

local _http_uri_router_table = {
    -- 兼容截图接口
    ['/api/screen/snapshot'] = function(ctx)
        if ctx.request_method ~= 'GET' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        ctx.status = 200
        local scr_w, scr_h = screen.size()
        local scale = _number_range(ctx.query['scale'], 1, 300, 100)
        local ltx = _number_range(ctx.query['ltx'], 0, scr_w - 1)
        local lty = _number_range(ctx.query['lty'], 0, scr_h - 1)
        local rbx = _number_range(ctx.query['rbx'], 0, scr_w - 1, scr_w - 1)
        local rby = _number_range(ctx.query['rby'], 0, scr_h - 1, scr_h - 1)
        local format = ctx.query['format'] or 'png'
        local function image_data_with_format(img, format)
            if format == 'jpg' or format == 'jpeg' then
                ctx:header('Content-type', 'image/jpeg')
                return img:jpeg_data()
            else
                ctx:header('Content-type', 'image/png')
                return img:png_data()
            end
        end
        if scale == 100 then
            if ltx == 0 and lty == 0 and rbx == 0 and rby == 0 then
                return image_data_with_format(screen.image(), format)
            else
                return image_data_with_format(screen.image(ltx, lty, rbx, rby), format)
            end
        else
            local zoom = scale / 100
            local img
            if ltx == 0 and lty == 0 and rbx == 0 and rby == 0 then
                img = screen.image()
            else
                img = screen.image(ltx, lty, rbx, rby)
            end
            local w, h = img:size()
            require('image.cv')
            return image_data_with_format(img:cv_resize(math.floor(w * zoom), math.floor(h * zoom)), format)
        end
    end,
    ['/api/config'] = function (ctx)
        -- 协议参考
        -- {
        --     "cloud": {
        --         "enable": true,
        --         "address": "ws://192.168.1.100:3000"
        --     },
        --     "notify_stop": false
        -- }
        ctx:header('Content-Type', 'application/json; charset=utf-8')
        if ctx.request_method == 'PUT' then
            ctx.status = 204
            local api_conf = json.decode(ctx.content)
            if type(api_conf) == 'table' then
                local conf = _read_conf()
                if type(api_conf.notify_stop) == 'boolean' then
                    conf.script_end_hint = api_conf.notify_stop
                end
                if type(api_conf.cloud) == 'table' and type(api_conf.cloud.enable) == 'boolean' and type(api_conf.cloud.address) == 'string' then
                    conf.open_cloud_control = api_conf.cloud
                end
                _write_conf(conf)
                noexecute.run_cmd(XXT_EXE_PATH, "dofile", XXT_BIN_PATH.."/launch.lua")
            end
            return ''
        elseif ctx.request_method == 'GET' then
            ctx.status = 200
            local conf = _read_conf()
            local api_conf = {}
            api_conf.notify_stop = conf.script_end_hint
            api_conf.cloud = type(conf.open_cloud_control) == 'table' and conf.open_cloud_control or {enable = false, address = 'ws://127.0.0.1:8888'}
            return json.encode(api_conf)
        end
        return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
    end,
    ['/api/system/log'] = function(ctx)
        ctx:header('Content-Type', 'text/html; charset=utf-8')
        if ctx.request_method == 'DELETE' then
            ctx.status = 204
            noexecute.rm_rf(XXT_LOG_PATH..'/sys.log')
            return ''
        elseif ctx.request_method == 'GET' then
            ctx.status = 200
            local log = file_lastline(XXT_LOG_PATH..'/sys.log', ctx.query['last'])
            return log
        end
        return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
    end,
    ['/api/system/respring'] = function(ctx)
        if ctx.request_method ~= 'POST' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        ctx:header('Content-Type', 'text/html; charset=utf-8')
        ctx.status = 204
        if get_openapi_func then
            get_openapi_func('/respring')(ctx)
        else
            xxtouch.post('/respring', '', '')
        end
        return ''
    end,
    ['/api/system/reboot'] = function(ctx)
        if ctx.request_method ~= 'POST' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        ctx:header('Content-Type', 'text/html; charset=utf-8')
        ctx.status = 204
        if get_openapi_func then
            get_openapi_func('/reboot2')(ctx)
        else
            xxtouch.post('/reboot2', '', '')
        end
        return ''
    end,
    ['/api/script'] = function(ctx)
        if ctx.request_method ~= 'GET' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        ctx.status = 200
        ctx:header('Content-type', 'application/json; charset=utf-8')
        local retlist = {}
        for name in lfs.dir(XXT_SCRIPTS_PATH) do
            if string.match(name, '.+%.lua') or string.match(name, '.+%.xxt') or string.match(name, '.+%.xpp') --[[or string.match(name, '.+%.tep')]] then
                retlist[#retlist + 1] = name
            end
        end
        return json.encode(retlist)
    end,
    ['/api/app/state'] = function(ctx)
        if ctx.request_method ~= 'GET' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        ctx.status = 200
        ctx:header('Content-type', 'application/json; charset=utf-8')
        local wifi_ip = "127.0.0.1"
        for i,v in ipairs(device.ifaddrs()) do
            if #tostring(v[1])>2 and tostring(v[1]):sub(1, 2)=="en" then
                wifi_ip = v[2]
                if tostring(v[1])=="en0" then
                    break
                end
            end
        end
        local conf = _read_conf()
        local w, h = screen.size()
        return json.encode{
            app = {
                version = sys.zeversion(),
                license = 'X.X.T.',
            },
            script = {
                select = conf['current_script'],
                running = lua_is_running(),
            },
            system = {
                scrw = w,
                scrh = h,
                os = 'ios',
                name = device.name(),
                sn = device.serial_number(),
                ndid = string.format("%08X-%016X", sys.MGCopyAnswer('ChipID'), sys.MGCopyAnswer('UniqueChipID')),
                udid = device.udid(),
                version = sys.version(),
                ip = wifi_ip,
                battery = device.battery_level(),
                log = file_lastline(XXT_LOG_PATH..'/sys.log'),
            }
        }
    end,
    ['/api/script/stop'] = function(ctx)
        if ctx.request_method ~= 'POST' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        ctx:header('Content-type', 'text/plain; charset=utf-8')
        if not get_openapi_func then
            local c, h, r = xxtouch.post('/stop_script')
            ctx.status = 204
            return r or ''
        else
            local ret = get_openapi_func('/stop_script')(ctx)
            ctx.status = 204
            return ret
        end
    end,
    ['/api/file'] = function(ctx)
        ctx.status = 200
        ctx:header('Content-type', 'application/octet-stream')
        if ctx.request_method == 'GET' then
            local path = ctx.query['path']
            path = type(path) == 'string' and path or '/'
            path = path:trim()
            path = XXT_HOME_PATH.."/"..path
            if lfs.attributes(path, 'mode') == 'directory' then
                local retlist = {}
                setmetatable(retlist, json.array_mt)
                for name in lfs.dir(path) do
                    if name~='.' and name~='..' then
                        local finfo = lfs.attributes(path..'/'..name)
                        if finfo then
                            if finfo.mode == 'directory' then
                                retlist[#retlist + 1] = {name = name, type = 'dir'}
                            else
                                retlist[#retlist + 1] = {name = name, type = 'file'}
                            end
                        end
                    end
                end
                return json.encode(retlist)
            else
                local f, errmsg = io.open(path, 'r')
                if f then
                    local s = f:read('*a')
                    f:close()
                    return s
                else
                    return _request_error_message(ctx, 404, string.format('%s: %s can not open file %q %s', ctx.uri, ctx.request_method, path, errmsg))
                end
            end
        elseif ctx.request_method == 'DELETE' then
            ctx.status = 204
            local path = ctx.query['path']
            path = type(path) == 'string' and path or ''
            path = path:trim()
            if path == '' or path == '/' then
                return _request_error_message(ctx, 404, string.format('%s: operation failed: %s %s %q', ctx.uri, ctx.request_method, 'invalid path', path))
            end
            path = XXT_HOME_PATH.."/"..path
            if lfs.attributes(path, 'mode') then
                noexecute.rm_rf(path)
                return ''
            else
                return _request_error_message(ctx, 404, string.format('%s: operation failed: %s %s %q', ctx.uri, ctx.request_method, 'invalid path', path))
            end
        elseif ctx.request_method == 'PUT' then
            ctx.status = 204
            local path = ctx.query['path']
            path = type(path) == 'string' and path or ''
            path = path:trim()
            if path == '' or path == '/' then
                return _request_error_message(ctx, 404, string.format('%s: operation failed: %s %s %q', ctx.uri, ctx.request_method, 'invalid path', path))
            end
            path = XXT_HOME_PATH.."/"..path
            if lfs.attributes(path, 'mode') == 'directory' then
                return _request_error_message(ctx, 404, string.format('%s: operation failed: %s %s %q', ctx.uri, ctx.request_method, 'invalid path', path))
            end
            local contentlen = tonumber(ctx.http_headers['Content-Length']) or 0
            if type(contentlen) == "number" and contentlen > 0 then
                local f, errmsg = io.open(path, 'w')
                if f then
                    f:write(ctx.content)
                    f:close()
                    sys.lchown(path, 501, 501)
                    return ''
                else
                    return _request_error_message(ctx, 404, string.format('%s: operation failed: %s %q %s', ctx.uri, ctx.request_method, path, errmsg))
                end
            else
                sys.mkdir_p(path)
                sys.lchown(path, 501, 501)
                return ''
            end
        end
        return _internal_error_message(ctx, 400, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
    end,
    ['/api/touch/down'] = function(ctx)
        if ctx.request_method ~= 'POST' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        local body = json.decode(ctx.content)
        body = type(body) == 'table' and body or {}
        local finger = math.floor(tonumber(body.finger) or 1)
        local x = math.floor(tonumber(body.x) or -1) or -1
        local y = math.floor(tonumber(body.y) or -1) or -1
        if finger >= 0 and finger <= 29 and x >= 0 and y >= 0 then
            touch.down(finger, x, y)
            ctx.status = 204
            ctx:header('Content-type', 'text/plain; charset=utf-8')
            return ''
        end
        return _request_error_message(ctx, 400, string.format('%s: operation failed: argument error', ctx.uri))
    end,
    ['/api/touch/move'] = function(ctx)
        if ctx.request_method ~= 'POST' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        local body = json.decode(ctx.content)
        body = type(body) == 'table' and body or {}
        local finger = math.floor(tonumber(body.finger) or 1)
        local x = math.floor(tonumber(body.x) or -1) or -1
        local y = math.floor(tonumber(body.y) or -1) or -1
        if finger >= 0 and finger <= 29 and x >= 0 and y >= 0 then
            touch.move(finger, x, y)
            ctx.status = 204
            ctx:header('Content-type', 'text/plain; charset=utf-8')
            return ''
        end
        return _request_error_message(ctx, 400, string.format('%s: operation failed: argument error', ctx.uri))
    end,
    ['/api/touch/up'] = function(ctx)
        if ctx.request_method ~= 'POST' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        local body = json.decode(ctx.content)
        body = type(body) == 'table' and body or {}
        local finger = math.floor(tonumber(body.finger) or 1)
        if finger >= 0 and finger <= 29 then
            touch.up(finger)
            ctx.status = 204
            ctx:header('Content-type', 'text/plain; charset=utf-8')
            return ''
        end
        return _request_error_message(ctx, 400, string.format('%s: operation failed: argument error', ctx.uri))
    end,
    ['/api/touch/tap'] = function(ctx)
        if ctx.request_method ~= 'POST' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        local body = json.decode(ctx.content)
        body = type(body) == 'table' and body or {}
        local finger = math.floor(tonumber(body.finger) or 1)
        local x = math.floor(tonumber(body.x) or -1) or -1
        local y = math.floor(tonumber(body.y) or -1) or -1
        if finger >= 0 and finger <= 29 and x >= 0 and y >= 0 then
            touch.down(finger, x, y)
            touch.up(finger)
            ctx.status = 204
            ctx:header('Content-type', 'text/plain; charset=utf-8')
            return ''
        end
        return _request_error_message(ctx, 400, string.format('%s: operation failed: argument error', ctx.uri))
    end,
    ['/api/key/down'] = function(ctx)
        if ctx.request_method ~= 'POST' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        local body = json.decode(ctx.content)
        body = type(body) == 'table' and body or {}
        local keycode = type(body.code) == 'string' and body.code or nil
        if keycode then
            if keycode:upper() == "HOME" then
                keycode = "HOMEBUTTON"
            end
            local ok, err = pcall(key.down, keycode)
            if not ok then
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, err))
            end
            ctx.status = 204
            ctx:header('Content-type', 'text/plain; charset=utf-8')
            return ''
        end
        return _request_error_message(ctx, 400, string.format('%s: operation failed: argument error', ctx.uri))
    end,
    ['/api/key/up'] = function(ctx)
        if ctx.request_method ~= 'POST' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        local body = json.decode(ctx.content)
        body = type(body) == 'table' and body or {}
        local keycode = type(body.code) == 'string' and body.code or nil
        if keycode then
            if keycode:upper() == "HOME" then
                keycode = "HOMEBUTTON"
            end
            local ok, err = pcall(key.up, keycode)
            if not ok then
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, err))
            end
            ctx.status = 204
            ctx:header('Content-type', 'text/plain; charset=utf-8')
            return ''
        end
        return _request_error_message(ctx, 400, string.format('%s: operation failed: argument error', ctx.uri))
    end,
    ['/api/key/press'] = function(ctx)
        if ctx.request_method ~= 'POST' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        local body = json.decode(ctx.content)
        body = type(body) == 'table' and body or {}
        local keycode = type(body.code) == 'string' and body.code or nil
        if keycode then
            if keycode:upper() == "HOME" then
                keycode = "HOMEBUTTON"
            end
            local ok, err = pcall(key.press, keycode)
            if not ok then
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, err))
            end
            ctx.status = 204
            ctx:header('Content-type', 'text/plain; charset=utf-8')
            return ''
        end
        return _request_error_message(ctx, 400, string.format('%s: operation failed: argument error', ctx.uri))
    end,
    ['/api/pasteboard/read'] = function(ctx)
        if ctx.request_method ~= 'GET' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        ctx.status = 200
        local data = pasteboard.read("public.plain-text")
        if data ~= "" then
            ctx:header('Content-type', 'text/plain; charset=utf-8')
            return data
        else
            data = pasteboard.read('public.png')
            if data ~= "" then
                ctx:header('Content-type', 'image/png')
                return data
            else
                data = pasteboard.read('public.jpeg')
                if data ~= "" then
                    ctx:header('Content-type', 'image/jpeg')
                    return data
                else
                    ctx:header('Content-type', 'application/json; charset=utf-8')
                    return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, 'supports copying plain text or plain images from devices only.'))
                end
            end
        end
    end,
    ['/api/pasteboard/write'] = function(ctx)
        if ctx.request_method ~= 'POST' then
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end
        ctx.status = 204
        ctx:header('Content-type', 'application/json; charset=utf-8')
        local body = json.decode(ctx.content)
        body = type(body) == 'table' and body or {}
        local uti = type(body.uti) == 'string' and body.uti or "public.plain-text"
        local data = type(body.data) == 'string' and body.data or nil
        if data then
            if uti == "public.plain-text" then
                pasteboard.write(data, uti)
                return ''
            elseif uti == "public.png" or uti == "public.jpeg" or uti == "public.image" then
                local imgdata = data:base64_decode()
                pasteboard.write(imgdata, uti)
                return ''
            else
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, 'transfer of clipboard contents other than plain text and plain images is not supported.'))
            end
        else
            return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, 'argument error.'))
        end
    end,
}

local _http_uri_router_match_table = {
    {   -- 兼容
        '/api/script/.+/run',
        function(ctx)
            if ctx.request_method ~= 'POST' then
                return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
            end
            ctx:header('Content-type', 'application/json; charset=utf-8')
            local path = string.match(ctx.uri, '/api/script/(.+)/run')
            if path then
                path = XXT_SCRIPTS_PATH.."/"..path
            else
                return _request_error_message(ctx, 404, string.format('%s: operation failed: %s', ctx.uri, 'invalid path'))
            end
            if lfs.attributes(path, 'mode') == 'directory' then
                return _request_error_message(ctx, 404, string.format('%s: path is directory: %s %q', ctx.uri, ctx.request_method, path))
            end
            local ext = path_manager.extension(path)
            local ok, to_path
            if ext == '.tep' or ext == '.tar' then
                ok, to_path = tar_extract(path)
                if ext == '.tar' then
                    os.remove(path)
                end
                if ok then
                    path = path_manager.join(to_path, 'main.lua')
                else
                    return _request_error_message(ctx, 500, string.format('%s: operation failed: %s', ctx.uri, to_path))
                end
            end
            if lfs.attributes(path, 'mode') == 'directory' then
                return _request_error_message(ctx, 404, string.format('%s: path is directory: %s %q', ctx.uri, ctx.request_method, path))
            end
            local fakectx = {}
            local ctxmeta = getmetatable(ctx)
            local debugger_ip = ctx.remote_address:split(':')
            table.remove(debugger_ip, #debugger_ip)
            debugger_ip = table.concat(debugger_ip, ':')
            local code
            if type(to_path) == 'string' then
                code = string.format(nLog_imp_str..';package.path=%q..";"..package.path;package.path=%q..";"..package.path;package.cpath=%q..";"..package.cpath;dofile(%q)', to_path.."?.xxt", to_path.."?.lua", to_path.."?.so", path)
            else
                code = string.format(nLog_imp_str..';dofile(%q)', path)
            end
            if not get_openapi_func then
                local c, _, r = xxtouch.post('/spawn', json.encode(ctx.http_headers), code)
                if c == 200 then
                    local ret = json.decode(r) 
                    ret = type(ret) == 'table' and ret or {code = 99, 'unknown error'}
                    if ret.code == 0 then
                        ctx.status = 204
                        return ''
                    end
                    ctx.status = 400
                else
                    ctx.status = c
                end
                return r
            end
            setmetatable(fakectx, {
                __index = function(obj, name)
                    if name == 'content' then
                        return code
                    elseif name == 'header' then
                        return function(obj, item) --[[return ctx:header(item)]] end
                    end
                    return ctxmeta.__index(ctx, name)
                end,
                __newindex = function(obj, name, value)
                    return ctxmeta.__newindex(ctx, name, value)
                end,
            })
            local lua_spawn = get_openapi_func('/spawn')
            if lua_spawn then
                local r, errmsg = lua_spawn(fakectx)
                if r then
                    local ret = json.decode(r)
                    ret = type(ret) == 'table' and ret or {code = 99, 'unknown error'}
                    if ret.code == 0 then
                        ctx.status = 204
                        return ''
                    end
                    ctx.status = 400
                    return r
                else
                    return _request_error_message(ctx, 404, string.format('%s: operation failed: %s %s', ctx.uri, ctx.request_method, errmsg))
                end
            else
                return _request_error_message(ctx, 404, string.format('%s: operation failed: %s %s', ctx.uri, ctx.request_method, '"/spawn" api not found.'))
            end
            -- local fakectx = {}
            -- local ctxmeta = getmetatable(ctx)
            -- setmetatable(fakectx, {
            --     __index = function(obj, name)
            --         if name == 'content' then
            --             return string.format('{"filename": %q}', path)
            --         elseif name == 'header' then
            --             return function(obj, item) --[[return ctx:header(item)]] end
            --         end
            --         return ctxmeta.__index(ctx, name)
            --     end,
            --     __newindex = function(obj, name, value)
            --         return ctxmeta.__newindex(ctx, name, value)
            --     end,
            -- })
            -- local launch_script_file = get_openapi_func('/launch_script_file')
            -- if launch_script_file then
            --     local ret, errmsg = launch_script_file(fakectx)
            --     if ret then
            --         ctx.status = 204
            --         return ret
            --     else
            --         return _request_error_message(ctx, 404, string.format('%s: operation failed: %s %s', ctx.uri, ctx.request_method, errmsg))
            --     end
            -- else
            --     return _request_error_message(ctx, 404, string.format('%s: operation failed: %s %s', ctx.uri, ctx.request_method, '"/launch_script_file" api not found.'))
            -- end
        end,
    },
    {   -- 兼容
        '^/api/script/.+/debug',
        function(ctx)
            if ctx.request_method ~= 'POST' then
                return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
            end
            ctx:header('Content-type', 'application/json; charset=utf-8')
            local path = string.match(ctx.uri, '^/api/script/(.+)/debug')
            if path then
                path = XXT_SCRIPTS_PATH.."/"..path
            else
                return _request_error_message(ctx, 404, string.format('%s: operation failed: %s', ctx.uri, 'invalid path'))
            end
            if lfs.attributes(path, 'mode') == 'directory' then
                return _request_error_message(ctx, 404, string.format('%s: path is directory: %s %q', ctx.uri, ctx.request_method, path))
            end
            local ext = path_manager.extension(path)
            local debugger_port = 8818
            local ok, to_path
            if ext == '.tep' or ext == '.tar' then
                ok, to_path = tar_extract(path)
                if ext == '.tar' then
                    os.remove(path)
                end
                if ok then
                    path = path_manager.join(to_path, 'main.lua')
                    local launch_json = file.reads(path_manager.join(to_path, '.vscode/launch.json'))
                    if launch_json then
                        local launch_json_tab = json.decode(launch_json)
                        if type(launch_json_tab) == 'table' and type(launch_json_tab.configurations) == 'table' and #(launch_json_tab.configurations) > 0 then
                            debugger_port = launch_json_tab.configurations[1].connectionPort
                        end
                    end
                else
                    return _request_error_message(ctx, 500, string.format('%s: operation failed: %s', ctx.uri, to_path))
                end
            end
            if lfs.attributes(path, 'mode') == 'directory' then
                return _request_error_message(ctx, 404, string.format('%s: path is directory: %s %q', ctx.uri, ctx.request_method, path))
            end
            local fakectx = {}
            local ctxmeta = getmetatable(ctx)
            local debugger_ip = ctx.remote_address:split(':')
            table.remove(debugger_ip, #debugger_ip)
            debugger_ip = table.concat(debugger_ip, ':')
            local code
            if type(to_path) == 'string' then
                code = string.format(nLog_imp_str..';package.path=%q..";"..package.path;package.path=%q..";"..package.path;package.cpath=%q..";"..package.cpath;require("LuaPanda").start(%q, %d);dofile(%q)', to_path.."?.xxt", to_path.."?.lua", to_path.."?.so", debugger_ip, debugger_port, path)
            else
                code = string.format(nLog_imp_str..';require("LuaPanda").start(%q, %d);dofile(%q)', debugger_ip, debugger_port, path)
            end
            sys.toast(string.format('调试主机: %s\n调试端口: %d\n调试脚本: %s', debugger_ip, debugger_port, path))
            if not get_openapi_func then
                local c, _, r = xxtouch.post('/spawn', json.encode(ctx.http_headers), code)
                if c == 200 then
                    local ret = json.decode(r) 
                    ret = type(ret) == 'table' and ret or {code = 99, 'unknown error'}
                    if ret.code == 0 then
                        ctx.status = 204
                        return ''
                    end
                    ctx.status = 400
                else
                    ctx.status = c
                end
                return r
            end
            setmetatable(fakectx, {
                __index = function(obj, name)
                    if name == 'content' then
                        return code
                    elseif name == 'header' then
                        return function(obj, item) --[[return ctx:header(item)]] end
                    end
                    return ctxmeta.__index(ctx, name)
                end,
                __newindex = function(obj, name, value)
                    return ctxmeta.__newindex(ctx, name, value)
                end,
            })
            local lua_spawn = get_openapi_func('/spawn')
            if lua_spawn then
                local r, errmsg = lua_spawn(fakectx)
                if r then
                    local ret = json.decode(r)
                    ret = type(ret) == 'table' and ret or {code = 99, 'unknown error'}
                    if ret.code == 0 then
                        ctx.status = 204
                        return ''
                    end
                    ctx.status = 400
                    return r
                else
                    return _request_error_message(ctx, 404, string.format('%s: operation failed: %s %s', ctx.uri, ctx.request_method, errmsg))
                end
            else
                return _request_error_message(ctx, 404, string.format('%s: operation failed: %s %s', ctx.uri, ctx.request_method, '"/spawn" api not found.'))
            end
        end,
    },
    {   -- 兼容
        '^/api/script/.+/encrypt',
        function(ctx)
            if ctx.request_method ~= 'POST' then
                return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
            end
            ctx:header('Content-type', 'application/json; charset=utf-8')
            return _request_error_message(ctx, 404, string.format('%s: operation failed: %s', ctx.uri, 'This feature is not available with this release'))
        end,
    },
    {   -- 兼容
        '^/api/script/.+',
        function(ctx)
            local path = string.match(ctx.uri, '^/api/script/(.+)')
            if path then
                path = XXT_SCRIPTS_PATH.."/"..path
            else
                return _request_error_message(ctx, 404, string.format('%s: operation failed: %s', ctx.uri, 'invalid path'))
            end
            if lfs.attributes(path, 'mode') == 'directory' then
                return _request_error_message(ctx, 404, string.format('%s: path is directory: %s %q', ctx.uri, ctx.request_method, path))
            end
            ctx:header('Content-type', 'text/plain; charset=utf-8')
            if ctx.request_method == 'DELETE' then
                local success, errmsg = os.remove(path)
                if success then
                    ctx.status = 204
                    return ''
                else
                    return _request_error_message(ctx, 404, string.format('%s: remove failed: %s %q %s', ctx.uri, ctx.request_method, path, errmsg))
                end
            elseif ctx.request_method == 'PUT' then
                local f, errmsg = io.open(path, 'w')
                if f then
                    f:write(ctx.content or '')
                    f:close()
                    sys.lchown(path, 501, 501)
                    ctx.status = 204
                    return ''
                else
                    return _request_error_message(ctx, 404, string.format('%s: write failed: %s %q %s', ctx.uri, ctx.request_method, path, errmsg))
                end
            elseif ctx.request_method == 'GET' then
                local f, errmsg = io.open(path, 'r')
                if f then
                    local s = f:read('*a')
                    f:close()
                    ctx.status = 200
                    return s
                else
                    return _request_error_message(ctx, 404, string.format('%s: read failed: %s %q %s', ctx.uri, ctx.request_method, path, errmsg))
                end
            end
            return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
        end,
    },
    {
        '^/api/path/jbroot/.-',
        function(ctx)
            if ctx.request_method ~= 'GET' then
                return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
            end
            ctx.status = 200
            ctx:header("Content-type", "text/plain; charset=utf-8")
            local path = string.match(ctx.uri, '^/api/path/jbroot(/.+)')
            if not path or path == '' then
                path = '/'
            end
            return jbroot(path)
        end,
    },
    {
        '^/api/path/rootfs/.-',
        function(ctx)
            if ctx.request_method ~= 'GET' then
                return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
            end
            ctx.status = 200
            ctx:header("Content-type", "text/plain; charset=utf-8")
            local path = string.match(ctx.uri, '^/api/path/rootfs(/.+)')
            if not path or path == '' then
                path = '/'
            end
            return rootfs(path)
		end,
	},
    {
        '^/api/proc%-value/.-',
        function(ctx)
            if ctx.request_method ~= 'POST' then
                return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
            end
            ctx:header('Content-type', 'application/json; charset=utf-8')
            local operation_name = string.match(ctx.uri, '^/api/proc%-value/(.+)')
            if type(operation_name) ~= 'string' or operation_name == '' then
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, 'operation not found.'))
            end
            local body = json.decode(ctx.content)
            if type(body) ~= "table" or "string" ~= type(body.key) then
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, 'argument error.'))
            end
            local operation = proc_value_operations[operation_name]
            if not operation then
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, 'operation not found.'))
            end
            local ret, err = operation(body)
            if ret then
                ctx.status = 200
                return json.encode(ret)
            else
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, err))
            end
        end,
    },
    {
        '^/api/proc%-queue/.-',
        function(ctx)
            if ctx.request_method ~= 'POST' then
                return _request_error_message(ctx, 405, string.format('%s: operation failed: method %s not allowed', ctx.uri, ctx.request_method))
            end
            ctx:header('Content-type', 'application/json; charset=utf-8')
            local operation_name = string.match(ctx.uri, '^/api/proc%-queue/(.+)')
            if type(operation_name) ~= 'string' or operation_name == '' then
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, 'operation not found.'))
            end
            local body = json.decode(ctx.content)
            if type(body) ~= "table" or "string" ~= type(body.key) then
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, 'argument error.'))
            end
            local operation = proc_queue_operations[operation_name]
            if not operation then
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, 'operation not found.'))
            end
            local ret, err= operation(body)
            if type(ret) == 'table' then
                ctx.status = 200
                return json.encode(ret)
            else
                return _request_error_message(ctx, 400, string.format('%s: operation failed: %s', ctx.uri, err))
            end
        end,
    },
}

local custom_http_api_factor, custom_http_api

local function _uri_router(ctx)

    local handler = _http_uri_router_table[ctx.uri]
    if type(handler) == 'function' then
        local noerr, ret = pcall(handler, ctx)
        if noerr then
            return ret
        else
            return _internal_error_message(ctx, 502, ret)
        end
    end

    for _,handler in ipairs(_http_uri_router_match_table) do
        if string.find(ctx.uri, handler[1]) == 1 then
            local noerr, ret = pcall(handler[2], ctx)
            if noerr then
                return ret
            else
                return _internal_error_message(ctx, 502, ret)
            end
        end
    end

    if not custom_http_api_factor then
		custom_http_api_factor = loadfile(XXT_BIN_PATH..'/custom-http-api.lua')
	end
    if type(custom_http_api_factor) == 'function' then
		local noerr = nil
		if not custom_http_api then
			noerr, custom_http_api = pcall(custom_http_api_factor)
		else
			noerr = true
		end
		if noerr then
			if type(custom_http_api) == 'function' then
				noerr, ret = pcall(custom_http_api, ctx)
				if noerr then
					return tostring(ret)
				else
					return _internal_error_message(ctx, 503, 'custom_http_api: '..tostring(ret))
				end
			end
		else
			return _internal_error_message(ctx, 503, 'custom_http_api_factor: '..tostring(custom_http_api))
		end
	end

end

return function (ctx)
    -- if ctx.uri ~= '/api/system/log' then
    --     sys.log(json.encode({uri = ctx.uri, method = ctx.request_method, headers = ctx.http_headers, query = ctx.query}))
    -- end
    local noerr, ret = pcall(_uri_router, ctx)
    if noerr then
        if type(ret) == 'string' then
            return ret
        end
    else
        return _internal_error_message(ctx, 501, ret)
    end
    ctx.status = 404
    ctx:header('Content-type', 'application/json; charset=utf-8')
    return [[{"code": 404, "message": "page not found."}]]
end